import argparse
import importlib_resources
from bin.cli_impl import *
from rich_argparse import RichHelpFormatter
from rich import print, inspect, print_json
from rich.rule import Rule
from rich.panel import Panel
from rich.padding import Padding
from rich.logging import RichHandler
from rich.console import Console
from rich.markdown import Markdown
from rich_argparse import RichHelpFormatter
from rich.traceback import install

install()
from bin import __version__ as citros_version


def parser_run(main_sub, epilog=None):
    parser = main_sub.add_parser(
        "run",
        description=Panel(
            Markdown(
                open(
                    importlib_resources.files(f"data.doc").joinpath("run.md"), "r"
                ).read()
            ),
        ),
        epilog=epilog,
        help="run section",
        formatter_class=RichHelpFormatter,
    )
    parser.add_argument("-dir", default=".", help="The working dir of the project")
    parser.add_argument(
        "-n", "--batch_name", nargs="?", default=None, help="a name for the run"
    )
    parser.add_argument(
        "-m", "--batch_message", nargs="?", default=None, help="a message for the run"
    )
    parser.add_argument(
        "-l",
        "--lan_traffic",
        action="store_true",
        help="receive LAN ROS traffic in your simulation.",
    )
    parser.add_argument(
        "-s", "--simulation_name", nargs="?", default=None, help="Simulation name"
    )
    parser.add_argument("-i", "--run_id", nargs="?", default="", help="run id")
    parser.add_argument(
        "-c",
        "--completions",
        nargs="?",
        default=1,
        help="number of times to run the simulation",
    )
    parser.add_argument(
        "-d", "--debug", action="store_true", help="set logging level to debug"
    )
    parser.add_argument(
        "-v", "--verbose", action="store_true", help="use verbose console prints"
    )
    parser.set_defaults(func=run)



import argparse
import importlib_resources
from bin.cli_impl import *
from rich_argparse import RichHelpFormatter
from rich import print, inspect, print_json
from rich.rule import Rule
from rich.panel import Panel
from rich.padding import Padding
from rich.logging import RichHandler
from rich.console import Console
from rich.markdown import Markdown
from rich_argparse import RichHelpFormatter
from rich.traceback import install
from bin import __version__ as citros_version

install()


# citros data list
def parser_data_list(parent_subparser, epilog=None):
    description_path = "data/list.md"
    help = "data list section"

    parser = parent_subparser.add_parser(
        "list",
        description=Panel(
            Markdown(
                open(
                    importlib_resources.files(f"data.doc").joinpath(description_path),
                    "r",
                ).read()
            ),
            subtitle=f"[{citros_version}]",
            title="description",
        ),
        epilog=epilog,
        help=help,
        formatter_class=RichHelpFormatter,
    )
    parser.add_argument("-dir", default=".", help="The working dir of the project")
    parser.add_argument(
        "-d", "--debug", action="store_true", help="set logging level to debug"
    )
    parser.add_argument(
        "-v", "--verbose", action="store_true", help="use verbose console prints"
    )
    parser.set_defaults(func=data_list)

    return parser


# citros data service status
def parser_data_service_status(parent_subparser, epilog=None):
    description_path = "data/service/status.md"
    help = "data service status section"

    parser = parent_subparser.add_parser(
        "service",
        description=Panel(
            Markdown(
                open(
                    importlib_resources.files(f"data.doc").joinpath(description_path),
                    "r",
                ).read()
            ),
            subtitle=f"[{citros_version}]",
            title="description",
        ),
        epilog=epilog,
        help=help,
        formatter_class=RichHelpFormatter,
    )

    parser.add_argument(
        "-d", "--debug", action="store_true", help="set logging level to debug"
    )
    parser.add_argument(
        "-v", "--verbose", action="store_true", help="use verbose console prints"
    )
    parser.set_defaults(func=data_service_status)

    subparser = parser.add_subparsers(dest="type")

    return parser


# citros data service
def parser_data_service(parent_subparser, epilog=None):
    description_path = "data/service.md"
    help = "data service section"

    parser = parent_subparser.add_parser(
        "service",
        description=Panel(
            Markdown(
                open(
                    importlib_resources.files(f"data.doc").joinpath(description_path),
                    "r",
                ).read()
            ),
            subtitle=f"[{citros_version}]",
            title="description",
        ),
        epilog=epilog,
        help=help,
        formatter_class=RichHelpFormatter,
    )
    parser.add_argument("-dir", default=".", help="The working dir of the project")
    parser.add_argument("-H", "--host", default="0.0.0.0", help="host")
    parser.add_argument("-p", "--port", default="8000", help="post to listen to")
    parser.add_argument("-t", "--time", action="store_true", help="print request times")
    parser.add_argument(
        "-d", "--debug", action="store_true", help="set logging level to debug"
    )
    parser.add_argument(
        "-v", "--verbose", action="store_true", help="use verbose console prints"
    )
    parser.set_defaults(func=data_service)

    subparser = parser.add_subparsers(dest="type")

    parser_data_service_status(subparser)

    return parser


# citros data db create
def parser_data_db_create(parent_subparser, epilog=None):
    description_path = "data/db/create.md"
    help = "data db create section"

    parser = parent_subparser.add_parser(
        "create",
        description=Panel(
            Markdown(
                open(
                    importlib_resources.files(f"data.doc").joinpath(description_path),
                    "r",
                ).read()
            ),
            subtitle=f"[{citros_version}]",
            title="description",
        ),
        epilog=epilog,
        help=help,
        formatter_class=RichHelpFormatter,
    )

    parser.add_argument(
        "-d", "--debug", action="store_true", help="set logging level to debug"
    )
    parser.add_argument(
        "-v", "--verbose", action="store_true", help="use verbose console prints"
    )
    parser.set_defaults(func=data_db_create)

    # subparser = parser.add_subparsers(dest="type")

    return parser


# citros data db status
def parser_data_db_status(parent_subparser, epilog=None):
    description_path = "data/db/status.md"
    help = "data db status section"

    parser = parent_subparser.add_parser(
        "status",
        description=Panel(
            Markdown(
                open(
                    importlib_resources.files(f"data.doc").joinpath(description_path),
                    "r",
                ).read()
            ),
            subtitle=f"[{citros_version}]",
            title="description",
        ),
        epilog=epilog,
        help=help,
        formatter_class=RichHelpFormatter,
    )

    parser.add_argument(
        "-d", "--debug", action="store_true", help="set logging level to debug"
    )
    parser.add_argument(
        "-v", "--verbose", action="store_true", help="use verbose console prints"
    )
    parser.set_defaults(func=data_db_status)

    # subparser = parser.add_subparsers(dest="type")

    return parser


# citros data db clean
def parser_data_db_clean(parent_subparser, epilog=None):
    description_path = "data/db/clean.md"
    help = "data db clean section"

    parser = parent_subparser.add_parser(
        "clean",
        description=Panel(
            Markdown(
                open(
                    importlib_resources.files(f"data.doc").joinpath(description_path),
                    "r",
                ).read()
            ),
            subtitle=f"[{citros_version}]",
            title="description",
        ),
        epilog=epilog,
        help=help,
        formatter_class=RichHelpFormatter,
    )

    parser.add_argument(
        "-d", "--debug", action="store_true", help="set logging level to debug"
    )
    parser.add_argument(
        "-v", "--verbose", action="store_true", help="use verbose console prints"
    )
    parser.set_defaults(func=data_db_clean)

    # subparser = parser.add_subparsers(dest="type")

    return parser


# citros data db
def parser_data_db(parent_subparser, epilog=None):
    description_path = "data/db.md"
    help = "data db section"

    parser = parent_subparser.add_parser(
        "db",
        description=Panel(
            Markdown(
                open(
                    importlib_resources.files(f"data.doc").joinpath(description_path),
                    "r",
                ).read()
            ),
            subtitle=f"[{citros_version}]",
            title="description",
        ),
        epilog=epilog,
        help=help,
        formatter_class=RichHelpFormatter,
    )
    parser.add_argument(
        "-d", "--debug", action="store_true", help="set logging level to debug"
    )
    parser.add_argument(
        "-v", "--verbose", action="store_true", help="use verbose console prints"
    )

    subparser = parser.add_subparsers(dest="type")

    parser_data_db_status(subparser)
    parser_data_db_clean(subparser)

    return parser


# citros data
def parser_data(main_sub, epilog=None):
    description_path = "data.md"
    help = "data section"

    parser = main_sub.add_parser(
        "data",
        description=Panel(
            Markdown(
                open(
                    importlib_resources.files(f"data.doc").joinpath(description_path),
                    "r",
                ).read()
            ),
            subtitle=f"[{citros_version}]",
            title="description",
        ),
        epilog=epilog,
        help=help,
        formatter_class=RichHelpFormatter,
    )
    parser.add_argument("-dir", default=".", help="The working dir of the project")
    parser.add_argument(
        "-d", "--debug", action="store_true", help="set logging level to debug"
    )
    parser.add_argument(
        "-v", "--verbose", action="store_true", help="use verbose console prints"
    )
    parser.set_defaults(func=data)

    subsubparser = parser.add_subparsers(dest="type")
    parser_data_list(subsubparser, epilog)
    parser_data_service(subsubparser, epilog)

    return parser



import os
import sys
import time
import json
import glob
import logging
from rich import print, inspect, print_json
from rich.panel import Panel
from rich.padding import Padding
from rich.logging import RichHandler
from watchdog.observers import Observer
from watchdog.events import FileSystemEventHandler, PatternMatchingEventHandler

# from batch import Batch
from citros.batch import Batch


# start listening on the folder changes and calls on_folder_change_handler on each change
class BatchRunFolderWatch:
    def __init__(self, path, log):
        self.observer = Observer()

        self.path = path
        self.log = log

        # we are looking on
        #   batch/info.json
        #   batch/simulation/info.json
        self.event_handler = HandlerPattern(
            root=self.path,
            log=self.log,
            patterns=["*/info.json"],
            ignore_patterns=[],
            ignore_directories=True,
        )

    def run(self):
        self.observer.schedule(self.event_handler, "tmp", recursive=True)

        self.observer.start()
        try:
            while True:
                time.sleep(5)
        except:
            self.observer.stop()
            self.log.exception("Error")

        self.observer.join()


class HandlerPattern(PatternMatchingEventHandler):
    def __init__(
        self,
        root,
        log=None,
        patterns=["*"],
        ignore_patterns=[],
        ignore_directories=False,
    ):
        PatternMatchingEventHandler.__init__(
            self,
            patterns=patterns,
            ignore_patterns=ignore_patterns,
            ignore_directories=ignore_directories,
        )
        self.root = root
        self.log = log

    def on_any_event(self, event):
        # inspect(event)
        # print("vova", self.root)
        # print("vova", event.src_path)
        # print("vova", event.src_path.removeprefix(self.root))
        # print("event.event_type", event.event_type)
        if event.event_type == "deleted":
            return
        simulation_name = (
            event.src_path.removeprefix(self.root).removeprefix("/").split("/")[0]
        )
        batch_name = (
            event.src_path.removeprefix(self.root).removeprefix("/").split("/")[1]
        )

        print(
            Panel(
                f"detected changed in simulaiton / batch run name: [green]{simulation_name}/{batch_name}[/green]"
            )
        )
        batch = Batch(self.root, simulation_name, batch_name, self.log)
        print_json(data=batch.data)


if __name__ == "__main__":
    logging.basicConfig(
        level=os.environ.get("LOGLEVEL", "INFO"),
        format="%(message)s",
        datefmt="[%X]",
        handlers=[RichHandler(rich_tracebacks=True)],
    )
    log = logging.getLogger(__name__)

    # path = sys.argv[1] if len(sys.argv) > 1 else '.'
    path = os.environ.get("RUNS_DIR", "tmp")

    print("[bold green]CITROS[/bold green] data-access")

    w = BatchRunFolderWatch(os.path.join(os.getcwd(), path), log)
    w.run()

    print("[bold green]CITROS[/bold green] data-access is out of the building.")



import argparse
import importlib_resources
from bin.cli_impl import *
from rich_argparse import RichHelpFormatter
from rich import print, inspect, print_json
from rich.rule import Rule
from rich.panel import Panel
from rich.padding import Padding
from rich.logging import RichHandler
from rich.console import Console
from rich.markdown import Markdown
from rich_argparse import RichHelpFormatter
from rich.traceback import install
from bin import __version__ as citros_version

from .parsers import (
    parser_run,
    parser_simulation,
    parser_parameter,
    parser_launch,
    parser_data,
    parser_report,
    parser_init,
    parser_doctor,
)

install()

# PANNEL = ""
PANNEL = Panel.fit(
    f"""[green]████████████████████████████████████████████████████████████
██      ███        ██        ██       ████      ████      ██
█  ████  █████  ████████  █████  ████  ██  ████  ██  ███████
█  ███████████  ████████  █████       ███  ████  ███      ██
█  ████  █████  ████████  █████  ███  ███  ████  ████████  █
██      ███        █████  █████  ████  ███      ████      ██
████████████████████████████████████████████████████████████""",
    subtitle=f"[{citros_version}]",
)
EPILOG = Markdown("Read more at [citros](https://citros.io)")


def main():
    # main parser -----------------------------------------
    parser = argparse.ArgumentParser(
        description=PANNEL,
        epilog=EPILOG,
        formatter_class=RichHelpFormatter,
        # add_help=False
    )
    parser.add_argument("-V", "--version", action="version", version=citros_version)

    subparsers = parser.add_subparsers(
        title="commands", help="citros commands", dest="type", required=True
    )

    parser_init(subparsers, EPILOG)
    parser_doctor(subparsers, EPILOG)
    parser_run(subparsers, EPILOG)
    parser_simulation(subparsers, EPILOG)
    parser_parameter(subparsers, EPILOG)
    parser_launch(subparsers, EPILOG)
    parser_data(subparsers, EPILOG)
    parser_report(subparsers, EPILOG)

    args, argv = parser.parse_known_args()

    args.func(args, argv)


if __name__ == "__main__":
    main()



import os
from pathlib import Path
from decouple import config as _conf


class config:
    CITROS_HOME_DIR: str = Path.home() / ".citros"
    """
    ~/.citros
    """

    ROOT_DIR: str = _conf("ROOT_DIR", None)
    """
    the directory where `.citros` is located.
    """

    RECORDINGS_DIR: str = _conf("RECORDINGS_DIR", None)
    """
    the directory where `citros run` records all data to
    """

    VERBOSE: bool = _conf("VERBOSE", default=False, cast=bool)
    """
    system wide vorbose variable
    """

    STATS_INTERVAL: int = _conf("STATS_INTERVAL", default=1, cast=int)
    """
    update interval for System Stats Recorder
    """

    CITROS_ENVIRONMENT: str = _conf("CITROS_ENVIRONMENT", "LOCAL")
    """
    where we are running, can be [LOCAL, CLOUD, GITHUB]
    """

    STORAGE_TYPE = _conf("STORAGE_TYPE", "MCAP")
    """the storage type that will be used to record the ros bag into. default is MCAP. 
    options: [SQLITE3, MCAP]
    """

    ROS_DOMAIN_ID: int = _conf("ROS_DOMAIN_ID", 42, cast=int)
    """
    the domain id for ros to use. 
    """

    OPEN_TELEMETRY_URL = _conf("OPEN_TELEMETRY_URL", "localhost:3417")
    """
    the url for opentelemetry
    """
    
    TRACE_CONTEXT = _conf("TRACE_CONTEXT", None)
    """
    if the trace was started on the cluster by a different entity (e.g. the worker),
    its context is given by an environment variable.
    """



import path
import sys
from citros import Citros
from pathlib import Path
from rich import print, inspect, print_json
from rich.rule import Rule
from rich.panel import Panel
from rich.padding import Padding
from rich.logging import RichHandler
from rich.console import Console
from rich.markdown import Markdown
from rich_argparse import RichHelpFormatter
from citros.utils import str_to_bool, suppress_ros_lan_traffic
from citros.batch import Batch
from rich.table import Table
from rich.console import Console
from rich import pretty
import json

pretty.install()


import glob
from .config import config

from citros import CitrosNotFoundException

directory = path.Path(__file__).abspath()
sys.path.append(directory.parent.parent)
from data_access import data_access as _data_access

from InquirerPy import prompt, inquirer
from prompt_toolkit.validation import Validator, ValidationError


class NumberValidator(Validator):
    """
    small helper class for validating user input during an interactive session.
    """

    def validate(self, document):
        try:
            int(document.text)
        except ValueError:
            raise ValidationError(
                message="Please enter a number", cursor_position=len(document.text)
            )


############################# CLI implementation ##############################
def init(args, argv):
    """
    :param args.dir:
    :param args.debug:
    :param args.verbose:
    """
    print(f'initializing CITROS at "{Path(args.dir).resolve()}". ')
    citros = Citros(new=True, root=args.dir, verbose=args.verbose, debug=args.debug)
    if args.debug:
        print("[green]done initializing CITROS")


def run(args, argv):
    """
    :param args.simulation_name:
    :param args.run_id:
    :param args.completions:

    :param args.batch_name:
    :param args.batch_message:

    :param args.lan_traffic:

    :param args.debug:
    :param args.verbose:
    """
    try:
        citros = Citros(root=args.dir, verbose=args.verbose, debug=args.debug)
    except CitrosNotFoundException:
        print(
            f'[red] "{Path(args.dir).expanduser().resolve()}" has not been initialized. cant run "citros run" on non initialized directory.'
        )
        return

    if args.debug:
        print("[green]done initializing CITROS")

    batch_name = args.batch_name
    batch_message = args.batch_message

    if not batch_name and str_to_bool(citros.settings["force_batch_name"]):
        print("[red]Please supply a batch name with flag -n <name>.")
        print(
            Panel.fit(
                Padding('You may run [green]"citros run -n <name>" ', 1), title="help"
            )
        )
        return False

    if not batch_message and str_to_bool(citros.settings["force_message"]):
        print("[red]Please supply a batch message with flag -m <message>.")
        print(
            Panel.fit(
                Padding('You may run [green]"citros run -m <message>"', 1), title="help"
            )
        )
        return False

    simulation = choose_simulation(
        citros,
        args.simulation_name,
    )

    root_rec_dir = f"{args.dir}/.citros/data"
    if config.RECORDINGS_DIR:
        root_rec_dir = config.RECORDINGS_DIR

    batch = Batch(
        root_rec_dir,
        simulation,
        name=batch_name,
        mesaage=batch_message,
        verbose=args.verbose,
        debug=args.debug,
    )
    batch.run(
        10, ros_domain_id=config.ROS_DOMAIN_ID, trace_context=config.TRACE_CONTEXT
    )


# helper function
def choose_simulation(citros: Citros, simulation_name):
    simulations_dict = {}
    for s in citros.simulations:
        simulations_dict[s.name] = s

    if simulation_name:
        return simulations_dict[simulation_name]
    sim_names = simulations_dict.keys()

    # sanity check - should never happen because internal_sync will fail if there
    #                isn't at least one simulation file.
    if not sim_names:
        print(
            f"[red]There are currently no simulations in your {citros.SIMS_DIR} folder. \
                	 Please create at least one simulation for your project."
        )
        return

    # interactive
    answers = prompt(
        [
            {
                "type": "list",
                "name": "sim_names",
                "message": "Please choose the simulation you wish to run:",
                "choices": sim_names,
            }
        ]
    )

    sim_name = answers.get("sim_names")
    return simulations_dict[sim_name]


def doctor(args, argv):
    # TODO[critical]: implement doctor
    print(f"[red] 'citros {args.func.__name__}' is Not implemented yet")


############################# Simulation implementation ##############################
def simulation_list(args, argv):
    # TODO[critical]: implement data_status
    print(f"[red] 'citros {args.func.__name__}' is Not implemented yet")


def simulation_run(args, argv):
    # TODO[critical]: implement data_status
    print(f"[red] 'citros {args.func.__name__}' is Not implemented yet")


############################# Simulation implementation ##############################
def parameter_setup_new(args, argv):
    # TODO[critical]: implement data_status
    print(f"[red] 'citros {args.func.__name__}' is Not implemented yet")


def parameter_setup_list(args, argv):
    # TODO[critical]: implement data_status
    print(f"[red] 'citros {args.func.__name__}' is Not implemented yet")


def parameter_setup(args, argv):
    # TODO[critical]: implement data_status
    print(f"[red] 'citros {args.func.__name__}' is Not implemented yet")


############################# DATA implementation ##############################
def data(args, argv):
    # -s simulation
    # -n name
    # -i version index
    # print batch info.
    root = Path(args.dir).expanduser().resolve() / ".citros/data"

    # simulation
    simulations_glob = sorted(glob.glob(f"{str(root)}/*"))
    simulations = []
    for sim in simulations_glob:
        if Path(sim).is_dir():
            simulations.append(str(sim).split("/")[-1])

    chosen_simulation = inquirer.fuzzy(
        message="Select Simulation:", choices=simulations, default="", border=True
    ).execute()

    # batch
    batch_glob = sorted(glob.glob(f"{str(root / chosen_simulation)}/*"))
    batches = []
    for batch in batch_glob:
        if Path(batch).is_dir():
            batches.append(str(batch).split("/")[-1])

    chosen_batch = inquirer.fuzzy(
        message="Select Batch:", choices=batches, default="", border=True
    ).execute()

    # version
    version_glob = sorted(glob.glob(f"{str(root / chosen_simulation/ chosen_batch)}/*"))
    versions = []
    for version in version_glob:
        if Path(version).is_dir():
            versions.append(str(version).split("/")[-1])

    version = inquirer.fuzzy(
        message="Select Version:", choices=versions, default="", border=True
    ).execute()

    root / chosen_simulation / chosen_batch / version

    batch = Batch(
        root,
        chosen_simulation,
        name=chosen_batch,
        debug=args.debug,
        verbose=args.verbose,
    )
    # inspect(batch)
    console = Console()
    console.rule(f"{chosen_simulation} / {chosen_batch} / {version}")
    console.print_json(data=batch.data)

    return


def data_list(args, argv):
    root = Path(args.dir).expanduser().resolve() / ".citros/data"

    table = Table(title="Simulation Runs")
    table.add_column("Simulation", style="cyan", no_wrap=True)
    table.add_column("Run name", style="magenta")
    table.add_column("Versions", justify="right", style="green")
    table.add_column("Data", justify="right", style="green")

    simulations = sorted(glob.glob(f"{str(root)}/*"))
    for sim in simulations:
        names = sorted(glob.glob(f"{sim}/*"))
        for name in names:
            versions = sorted(glob.glob(f"{name}/*"))
            for version in versions:
                table.add_row(
                    sim.split("/")[-1],
                    name.split("/")[-1],
                    version.split("/")[-1],
                    "vova",
                )

    console = Console()
    console.print(table)


def data_service(args, argv):
    """
    :param args.dir
    :param args.debug:
    :param args.verbose:
    :param args.project_name:
    """

    root = Path(args.dir).expanduser().resolve() / ".citros/data"
    print(
        Panel.fit(
            f"""started at [green]http://{args.host}:{args.port}[/green].
API: open [green]http://{args.host}:{args.port}/redoc[/green] for documantation
Listening on: [green]{str(root)}""",
            title="[green]CITROS service",
        )
    )

    _data_access(
        str(root),
        time=args.time,
        host=args.host,
        port=int(args.port),
        debug=args.debug,
        verbose=args.verbose,
    )


def data_service_status(args, argv):
    # TODO[critical]: implement data_status
    print(f"[red] 'citros {args.func.__name__}' is Not implemented yet")


def data_db_create(args, argv):
    # TODO[critical]: implement data_status
    print(f"[red] 'citros {args.func.__name__}' is Not implemented yet")


def data_db_status(args, argv):
    # TODO[critical]: implement data_create
    print(f"[red] 'citros {args.func.__name__}' is Not implemented yet")


def data_db_clean(args, argv):
    # TODO[critical]: implement data_clean
    print(f"[red] 'citros {args.func.__name__}' is Not implemented yet")


############################# REPORT implementation ##############################
def report_generate(args, argv):
    # TODO[critical]: implement report_generate
    print(f"[red] 'citros {args.func.__name__}' is Not implemented yet")


def report_validate(args, argv):
    # TODO[critical]: implement report_validate
    print(f"[red] 'citros {args.func.__name__}' is Not implemented yet")



from unittest import TestCase

from tests.test_parse_makefile import *


class Test(TestCase):
    def test_test_px4_makefile(self):
        test_px4_makefile()

    def test_test_parse_makefile_no_install(self):
        test_parse_makefile_no_install()

    def test_test_micro_ros_agent_makefile(self):
        test_micro_ros_agent_makefile()



import os
from decouple import config as _conf


class config:
    # CITROS_DIR: str = _conf("CITROS_DIR", None)
    # """
    # the directory where `.citros` is located.
    # """

    ROOT_DIR: str = _conf(
        "ROOT_DIR", default=os.path.join(os.getcwd(), ".citros/data"), cast=str
    )
    """
    the directory where `citros run` records all data to
    """



import os
import json
import sys
import uuid
import shutil
from pathlib import Path
from citros.stats import SystemStatsRecorder
from rich import print, inspect, print_json
from rich.rule import Rule
from rich.panel import Panel
from rich.padding import Padding
from rich.logging import RichHandler
from rich.console import Console
from rich.markdown import Markdown
from rich_argparse import RichHelpFormatter
from datetime import datetime

from .logger import get_logger, shutdown_log

from .utils import suppress_ros_lan_traffic, validate_dir
from .parameter_setup import ParameterSetup
from .citros_obj import (
    CitrosObj,
    CitrosException,
    CitrosNotFoundException,
    FileNotFoundException,
    NoValidException,
)
from .events import EventsOTLP


class Simulation(CitrosObj):
    """Object representing .citros/simulations/name.json file."""

    ##################
    ##### public #####
    ##################
    def _get_simulation_run_log(self, simulation_rec_dir):
        return get_logger(
            __name__,
            log_level=os.environ.get("LOGLEVEL", "DEBUG" if self.debug else "INFO"),
            log_file=str(simulation_rec_dir / "citros.log"),
            verbose=self.verbose,
        )

    def copy_ros_log(self, destination: str):
        self.log.debug(f"{'   '*self.level}{self.__class__.__name__}.copy_ros_log()")
        from .utils import get_last_created_file, copy_files, rename_file

        ros_logs_dir_path = get_last_created_file(
            Path("~/.ros/log/").expanduser(), dirs=True
        )
        if get_last_created_file is None:
            self.log.warning(f"Failed to find the ros logs directory.")
            return

        log_file_path = Path(ros_logs_dir_path, "launch.log")
        copy_files([log_file_path], destination, self.log)
        new_file_path = Path(destination, log_file_path.name)
        rename_file(new_file_path, "ros.log")

    def copy_msg_files(self, destination: str):
        self.log.debug(f"{'   '*self.level}{self.__class__.__name__}.copy_msg_files()")
        from .utils import copy_files
        from .parsers import ParserRos2

        msg_paths = ParserRos2(self.log).get_msg_files(self.root)
        for msg_path in msg_paths:
            # assuming msg files are under package_name/msg/
            package_name = Path(msg_path).parent.parent.name
            target_dir = Path(destination, package_name, "msg")
            copy_files([msg_path], str(target_dir), self.log, True)

    def save_system_vars(self, destination: str):
        import subprocess
        from os import linesep

        self.log.debug(
            f"{'   '*self.level}{self.__class__.__name__}.save_system_vars()"
        )
        # Get all environment variables
        env_vars = dict(os.environ)

        pip_freeze_output = subprocess.run(
            ["pip", "freeze"], capture_output=True, text=True
        )

        if pip_freeze_output.returncode != 0:
            self.log.error("pip freeze failed: " + pip_freeze_output.stderr)
            python_packages = []
        else:
            python_packages = pip_freeze_output.stdout.split(linesep)

        data = {"environment_variables": env_vars, "python_packages": python_packages}

        with open(Path(destination, "environment.json"), "w") as f:
            json.dump(data, f, indent=4)

    def run(self, simulation_rec_dir, trace_context=None, ros_domain_id=None):
        """Run simulation."""
        # create .citros/data if not exists
        simulation_rec_dir.mkdir(parents=True, exist_ok=True)

        self.log = self._get_simulation_run_log(simulation_rec_dir)

        self.log.debug(f"{'   '*self.level}{self.__class__.__name__}.run()")
        events = EventsOTLP(self, trace_context)

        # running inside ROS workspace context.
        from launch import LaunchService
        from citros.ros import generate_launch_description

        self.log.info(f"running simulation [{self.name}]")
        # print(f"running simulation [{self.name}]")

        if self.verbose:
            self.log.info(f'simulation run dir = "{simulation_rec_dir}]"')
        else:
            self.log.debug(f'simulation run dir = "{simulation_rec_dir}]"')

        if ros_domain_id:
            suppress_ros_lan_traffic(ros_domain_id)

        # launch
        launch_description = generate_launch_description(
            self, simulation_rec_dir, events
        )

        if launch_description is None:
            msg = f"ERROR. Failed to create launch_description."
            self.log.error(msg)
            return

        launch_service = LaunchService(debug=False)  # self.debug)
        launch_service.include_launch_description(launch_description)

        systemStatsRecorder = SystemStatsRecorder(f"{simulation_rec_dir}/stats.csv")
        systemStatsRecorder.start()

        ret = launch_service.run()

        systemStatsRecorder.stop()

        print(
            f"[{'blue' if ret == 0 else 'red'}] - - Finished simulation with return code [{ret}].",
        )

        self.copy_ros_log(simulation_rec_dir)
        self.copy_msg_files(simulation_rec_dir)
        self.save_system_vars(simulation_rec_dir)

        # TODO!
        # self.save_run_data()

        if ret != 0:
            events.error(
                message=f"Finished simulation. Return code = [{ret}].",
            )
            events.on_shutdown()
            # sys.exit(ret)
        else:
            events.done(
                message=f"Finished simulation. Return code = [{ret}].",
            )
        return ret

    # overriding
    def path(self):
        return self.root_citros / "simulations" / f"{self.file}"

    ###################
    ##### private #####
    ###################

    def __init__(
        self,
        name,
        root=None,
        new=False,
        log=None,
        citros=None,
        package_name=None,
        launch_file=None,
        verbose=False,
        debug=False,
        level=0,
    ):
        # used for new simulation
        self.package_name = package_name
        self.launch_file = launch_file
        super().__init__(name, root, new, log, citros, verbose, debug, level)

    # overriding
    def _validate(self):
        self.log.debug(f"{'   '*self.level}{self.__class__.__name__}._validate()")

        """Validate simulation.json file."""
        Path(self.root).mkdir(parents=True, exist_ok=True)

        # validate json schema is correct
        success = validate_dir(
            self.root_citros / "simulations", "schema_simulation.json", self.log
        )

        if not success:
            self.log.debug(f"{'   '*self.level}{self.__class__.__name__}: False")
            return False

        # validate parameter_setup file
        param_setup = self["parameter_setup"]

        parameter_setup = ParameterSetup(
            param_setup,
            root=self.root,
            new=self.new,
            log=self.log,
            citros=self.citros,
            verbose=self.verbose,
            debug=self.debug,
            level=self.level + 1,
        )

        # validate launch file
        launch_file = self["launch"]["file"]
        all_launch_names = [launch["name"] for launch in self._get_launches()]
        if launch_file not in all_launch_names:
            print(
                f'[red]Could not find launch file named {launch_file} referenced in "{self.path()}."',
            )
            self.log.debug(f"{'   '*self.level}{self.__class__.__name__}: return False")
            return False

        self.log.debug(f"{'   '*self.level}{self.__class__.__name__}: return True")
        return True

    # overriding
    def _load(self):
        self.log.debug(f"{'   '*self.level}{self.__class__.__name__}._load()")

        try:
            # loads self.path()
            super()._load()
        except FileNotFoundError as ex:
            self.log.error(f"simulation file {self.file} does not exist.")
            raise FileNotFoundException(f"simulation file {self.file} does not exist.")

    # overriding
    def _new(self):
        self.log.debug(f"{'   '*self.level}{self.__class__.__name__}._new()")
        path = self.path()

        # avoid overwrite
        if path.exists():
            self._load()
            # return

        Path(self.root_citros / "simulations").mkdir(parents=True, exist_ok=True)

        # got from __init__
        if self.launch_file is None or self.package_name is None:
            raise ValueError(
                "package_name and launch_file must be provided when creating a new simulation"
            )

        default = {
            "id": str(uuid.uuid4()),
            "description": "Default simulation. Change the values according to your needs.",
            "parameter_setup": "default_param_setup.json",
            "launch": {"file": self.launch_file, "package": self.package_name},
            "timeout": 60,
            "GPU": 0,
            "CPU": 2,
            "MEM": 265,
            "storage_type": "MCAP",
        }
        self.data = default | self.data
        self._save()

    ###################
    ##### utils #####
    ###################

    def _get_launches(self):
        """returns a list of launch objects

        Args:
            proj_json (Path): path to project.json file

        Returns:
            [{
                package: str,
                name: str
            }]: array of launch info
        """

        launch_info = []

        for package in self.citros.get("packages", []):
            for launch in package.get("launches", []):
                if "name" in launch:
                    launch_info.append(
                        {"package": package.get("name", ""), "name": launch["name"]}
                    )

        return launch_info

    def _check_simualtion_run_name(self, name):
        batch_name_idx = 1

        if not name or not self.utils.is_valid_file_name(name):
            name = self.utils.get_foramtted_datetime()

        # avoid duplicate batch dir names
        elif Path(
            self._router()["citros"]["runs"]["path"],
            self._sim_name,
            name,
        ).exists():
            while Path(
                self._router()["citros"]["runs"]["path"],
                self._sim_name,
                f"{name}_{str(batch_name_idx)}",
            ).exists():
                batch_name_idx = batch_name_idx + 1
            name = f"{name}_{str(batch_name_idx)}"
        return name



import json
import shutil

from pathlib import Path

import importlib_resources
from citros.parsers import ParserRos2
from .citros_obj import CitrosObj
from .utils import get_data_file_path, validate_dir


class ParameterSetup(CitrosObj):
    """Object representing .citros/simulation.json file."""

    ##################
    ##### public #####
    ##################

    # overriding
    def path(self):
        return self.root_citros / "parameter_setups" / f"{self.file}"

    ###################
    ##### private #####
    ###################

    # overriding
    def _validate(self):
        """Validate parameter_setup.json file."""
        self.log.debug(f"{'   '*self.level}{self.__class__.__name__}._validate()")

        success = validate_dir(
            self.root_citros / "parameter_setups", "schema_param_setup.json", self.log
        )

        return success

    def _new(self):
        self.log.debug(f"{'   '*self.level}{self.__class__.__name__}._new()")
        path = self.path()

        # avoid overwrite
        if path.exists():
            self._load()
            # return

        Path(self.root_citros / "parameter_setups").mkdir(parents=True, exist_ok=True)

        self.parser_ros2 = ParserRos2(self.log)
        # get default parameter setup and write it to file.
        default_parameter_setup = self.parser_ros2.generate_default_params_setup(
            self.citros
        )
        self.data = default_parameter_setup | self.data
        self._save()

        # copy functions
        destination = self.root_citros / "parameter_setups" / "functions"
        destination.mkdir(parents=True, exist_ok=True)

        if not (destination / "my_func.py").exists():  # avoid overwriting
            importlib_resources.files(f"data.sample_code").joinpath("my_func.py")

            with importlib_resources.files(f"data.sample_code").joinpath(
                "my_func.py"
            ) as md_file_path:
                shutil.copy2(md_file_path, destination / f"my_func.py")

            with importlib_resources.files(f"data.doc").joinpath(
                "parameter_setups/functions/README.md"
            ) as md_file_path:
                shutil.copy2(md_file_path, destination / f"README.md")

        # TODO: when addeing a parameter to a package/node, add/append it to the parameter setups that use it.



import pytest
import numpy
from pathlib import Path
from tmp.citros.citros.ros.params import citros_params


class MockCitros:
    def __init__(self):
        self.log = self
        self.PARAMS_FUNCTIONS_DIR = Path("input/param_functions")

    # in case we use self.debug in citros_params
    def debug(self, msg):
        print(msg)


@pytest.fixture
def citros_params_obj():
    citros = MockCitros()
    return citros_params(citros)


def test_evaluate_numpy_function(citros_params_obj):
    data = {"math_op": {"function": "numpy.add", "args": [5, 3]}}
    result = citros_params_obj.evaluate_dictionary(data)
    assert result == {"math_op": 8}


def test_evaluate_user_function1(citros_params_obj):
    # Ensure you have a function in your specified directory for this test
    data = {"custom_math": {"function": "custom_math.py:addition", "args": [5, 3]}}
    result = citros_params_obj.evaluate_dictionary(data)
    assert result == {"custom_math": 8}


def test_evaluate_user_function2(citros_params_obj):
    # Ensure you have a function in your specified directory for this test
    data = {
        "custom_math": {"function": "custom_math.py:addition_with_np", "args": [5, 3]}
    }
    result = citros_params_obj.evaluate_dictionary(data)
    assert result == {"custom_math": 8}


def test_recursive_evaluation(citros_params_obj):
    data = {
        "a": 10,
        "b": {"function": "numpy.add", "args": ["a", 5]},
        "c": {"function": "numpy.multiply", "args": ["b", 2]},
    }
    result = citros_params_obj.evaluate_dictionary(data)
    assert result == {"a": 10, "b": 15, "c": 30}


def test_circular_dependency_detection(citros_params_obj):
    data = {
        "a": {"function": "numpy.add", "args": ["b", 1]},
        "b": {"function": "numpy.add", "args": ["a", 1]},
    }
    with pytest.raises(ValueError, match="Circular dependency detected"):
        citros_params_obj.evaluate_dictionary(data)


def test_nested_key_reference_multi_level(citros_params_obj):
    data = {
        "outer": {
            "inner_a": 5,
            "inner_b": {"function": "numpy.add", "args": ["inner_a", 3]},
        },
        "sum": {"function": "numpy.add", "args": ["outer.inner_b", 2]},
    }
    result = citros_params_obj.evaluate_dictionary(data)
    assert result == {"outer": {"inner_a": 5, "inner_b": 8}, "sum": 10}


def test_nested_key_reference_single_level(citros_params_obj):
    data = {
        "outer": {
            "inner_a": 5,
            "inner_b": {"function": "numpy.add", "args": ["inner_a", 3]},
        },
        "sum": {"function": "numpy.add", "args": ["inner_b", 2]},
    }
    result = citros_params_obj.evaluate_dictionary(data)
    assert result == {"outer": {"inner_a": 5, "inner_b": 8}, "sum": 10}


def test_list_processing(citros_params_obj):
    data = {"list": [1, 2, 3], "sum": {"function": "numpy.sum", "args": ["list"]}}
    result = citros_params_obj.evaluate_dictionary(data)
    assert result == {"list": [1, 2, 3], "sum": 6}


def test_recursive_evaluation_multiple_files(citros_params_obj):
    data = {
        "d": {"function": "more_custom_math.py:multiply_with_np", "args": ["c", "b"]},
        "a": 10,
        "b": {"function": "custom_math.py:addition", "args": ["a", 5]},
        "c": {"function": "custom_math.py:addition_with_np", "args": ["b", 2]},
    }
    result = citros_params_obj.evaluate_dictionary(data)
    assert result == {"a": 10, "b": 15, "c": 17, "d": 255}


def test_numpy_normal(citros_params_obj):
    data = {
        "mu": 0.0,
        "sigma": 1.0,
        "random_normal": {"function": "numpy.random.normal", "args": ["mu", "sigma"]},
    }
    result = citros_params_obj.evaluate_dictionary(data)
    assert isinstance(result["random_normal"], float)


def test_numpy_uniform(citros_params_obj):
    data = {
        "low": 1.0,
        "high": 10.0,
        "random_uniform": {"function": "numpy.random.uniform", "args": ["low", "high"]},
    }
    result = citros_params_obj.evaluate_dictionary(data)
    assert 1.0 <= result["random_uniform"] < 10.0


def test_numpy_exponential(citros_params_obj):
    data = {
        "scale": 2.0,
        "random_exponential": {
            "function": "numpy.random.exponential",
            "args": ["scale"],
        },
    }
    result = citros_params_obj.evaluate_dictionary(data)
    assert result["random_exponential"] >= 0



import argparse
import importlib_resources
from bin.cli_impl import *
from rich_argparse import RichHelpFormatter
from rich import print, inspect, print_json
from rich.rule import Rule
from rich.panel import Panel
from rich.padding import Padding
from rich.logging import RichHandler
from rich.console import Console
from rich.markdown import Markdown
from rich_argparse import RichHelpFormatter
from rich.traceback import install
from bin import __version__ as citros_version

install()


# citros parameter setup new
def parser_parameter_setup_new(subparser, epilog=None):
    description_path = "parameter/setup/new.md"
    help = "citros parameter new section"

    parameter_setup_subparser = subparser.add_parser(
        "new",
        description=Panel(
            Markdown(
                open(
                    importlib_resources.files(f"data.doc").joinpath(description_path),
                    "r",
                ).read()
            ),
            subtitle=f"[{citros_version}]",
            title="description",
        ),
        epilog=epilog,
        help=help,
        formatter_class=RichHelpFormatter,
    )
    parameter_setup_subparser.add_argument(
        "-d", "--debug", action="store_true", help="set logging level to debug"
    )
    parameter_setup_subparser.add_argument(
        "-v", "--verbose", action="store_true", help="use verbose console prints"
    )
    parameter_setup_subparser.set_defaults(func=parameter_setup_new)
    return parameter_setup_subparser


# citros parameter setup list
def parser_parameter_setup_list(subparser, epilog=None):
    description_path = "parameter/setup/list.md"
    help = "citros parameter setup list section"

    parameter_setup_subparser = subparser.add_parser(
        "list",
        description=Panel(
            Markdown(
                open(
                    importlib_resources.files(f"data.doc").joinpath(description_path),
                    "r",
                ).read()
            ),
            subtitle=f"[{citros_version}]",
            title="description",
        ),
        epilog=epilog,
        help=help,
        formatter_class=RichHelpFormatter,
    )
    parameter_setup_subparser.add_argument(
        "-d", "--debug", action="store_true", help="set logging level to debug"
    )
    parameter_setup_subparser.add_argument(
        "-v", "--verbose", action="store_true", help="use verbose console prints"
    )
    parameter_setup_subparser.set_defaults(func=parameter_setup_list)
    return parameter_setup_subparser


# citros parameter setup
def parser_parameter_setup(subparser, epilog=None):
    description_path = "parameter/setup.md"
    help = "citros data create section"

    parser = subparser.add_parser(
        "setup",
        description=Panel(
            Markdown(
                open(
                    importlib_resources.files(f"data.doc").joinpath(description_path),
                    "r",
                ).read()
            ),
            subtitle=f"[{citros_version}]",
            title="description",
        ),
        epilog=epilog,
        help=help,
        formatter_class=RichHelpFormatter,
    )
    parser.add_argument(
        "-d", "--debug", action="store_true", help="set logging level to debug"
    )
    parser.add_argument(
        "-v", "--verbose", action="store_true", help="use verbose console prints"
    )
    parser.set_defaults(func=parameter_setup)

    subsubparser = parser.add_subparsers(dest="type")

    parser_parameter_setup_new(subsubparser)
    parser_parameter_setup_list(subsubparser)

    return parser


# citros parameter
def parser_parameter(subparser, epilog=None):
    description_path = "parameter.md"
    help = "simulation section"

    parser = subparser.add_parser(
        "parameter",
        description=Panel(
            Markdown(
                open(
                    importlib_resources.files(f"data.doc").joinpath(description_path),
                    "r",
                ).read()
            ),
            subtitle=f"[{citros_version}]",
            title="description",
        ),
        epilog=epilog,
        help=help,
        formatter_class=RichHelpFormatter,
        aliases=["param"],
    )
    parser.add_argument("-n", "--name", default=None, help="name of parameter")
    parser.add_argument("--package", default=None, help="ilter package")
    parser.add_argument("--node", default=None, help="filter node")
    parser.add_argument("-m", "--match", default=None, help="match parameter pattern")

    subparser = parser.add_subparsers(dest="type")

    # simulation run/list
    parser_parameter_setup(subparser, epilog=epilog)

    return parser



import json
import glob
import shutil
import importlib_resources

from os import linesep
from pathlib import Path

from citros.parsers import ParserRos2
from citros.utils import validate_dir, validate_file

# from .ros import Ros
from .settings import Settings
from .simulation import Simulation
from .parameter_setup import ParameterSetup
from .citros_obj import (
    CitrosObj,
    CitrosException,
    FileNotFoundException,
    CitrosNotFoundException,
    NoValidException,
)

from rich.traceback import install
from rich.logging import RichHandler
from rich import print, inspect, print_json
from rich.panel import Panel
from rich.padding import Padding

install()


class Citros(CitrosObj):
    """Object representing .citros/simulations/{name}.json file."""

    # def __enter__(self):
    #     """
    #     Returns the Citros instance. This allows the class to be used in a `with` statement.
    #     """
    #     return self

    # def __exit__(self, exc_type, exc_val, exc_tb):
    #     """
    #     Makes sure the stats collecting thread is stopped and handles exceptions.

    #     Args:
    #     exc_type: The type of exception.
    #     exc_val: The exception instance.
    #     exc_tb: A traceback object encapsulating the call stack at the point
    #             where the exception originally occurred.
    #     """
    #     self.events.on_shutdown()

    #     self.systemStatsRecorder.stop()

    #     if exc_type is not None:
    #         self._handle_exceptions(exc_val, exit=True)

    def __init__(
        self,
        name="project",
        root=None,
        new=False,
        log=None,
        citros=None,
        verbose=False,
        debug=False,
        level=0,
    ):
        ###################
        ##### .citros #####
        ###################
        # init settings
        self.settings = None

        # init parameter_setups
        self.parameter_setups = []

        # init simulations
        self.simulations = []

        #################
        ##### utils #####
        #################
        self._ros = None

        super().__init__(name, root, new, log, citros, verbose, debug, level)

    def __str__(self):
        # print_json(data=self.data)
        return json.dumps(self.data, indent=4)

    ###################
    ##### private #####
    ###################
    # overriding
    def _validate(self):
        """Validate the json file."""

        # TODO: check that the project.json file is valid

        return True

    # overriding
    def _load(self):
        self.log.debug(f"{'   '*self.level}{self.__class__.__name__}.load()")

        # loads the main file (project.json)
        try:
            self.log.debug(f"loading .citros/project.json")
            super()._load()
        except FileNotFoundError as ex:
            self.log.error(f"simulation file {self.file} does not exist.")
            raise FileNotFoundException(f"simulation file {self.file} does not exist.")

        # loads the settings.json file
        self.log.debug(f"loading .citros/settings.json")
        self.settings = Settings(
            "settings",
            root=self.root,
            log=self.log,
            citros=self,
            new=self.new,
            debug=self.debug,
            verbose=self.verbose,
            level=self.level + 1,
        )

        # loads the parameter_setups
        for file in glob.glob(
            "*.json", root_dir=f"{self.root_citros}/parameter_setups/"
        ):
            self.log.debug(f"loading parameter_setup: {file}")
            self.parameter_setups.append(
                ParameterSetup(
                    file,
                    root=self.root,
                    new=self.new,
                    log=self.log,
                    citros=self,
                    debug=self.debug,
                    verbose=self.verbose,
                    level=self.level + 1,
                )
            )

        # loads the simulations
        for file in glob.glob("*.json", root_dir=f"{self.root_citros}/simulations/"):
            # self.simulations.append(Simulation(self.root, file, self.log, citros=self))
            self.log.debug(f"loading simulation: {file}")
            self.simulations.append(
                Simulation(
                    file,
                    root=self.root,
                    new=self.new,
                    log=self.log,
                    citros=self,
                    debug=self.debug,
                    verbose=self.verbose,
                    level=self.level + 1,
                )
            )

        # utils
        # self.ros = Ros(self.root, "ros.json", self.log)

    # overriding
    def _new(self):
        self.log.debug(f"{'   '*self.level}{self.__class__.__name__}._new()")

        # create the .citros folder
        Path(self.root_citros).mkdir(parents=True, exist_ok=True)

        # settings
        self.log.debug(f"creating .citros/settings.json")
        self.settings = Settings(
            "settings",
            root=self.root,
            log=self.log,
            citros=self,
            new=self.new,
            verbose=self.verbose,
            debug=self.debug,
            level=self.level + 1,
        )

        self._parser_ros2 = ParserRos2(self.log)
        # get data from ros2
        project_data = self._parser_ros2.parse(str(self.root))
        with open(self.path(), "w") as file:
            json.dump(project_data, file, sort_keys=True, indent=4)

        self.data = project_data
        self._save()

        # parameter setups
        self.log.debug(f"creating .citros/parameter_setups/default_param_setup.json")
        self.parameter_setups = ParameterSetup(
            "default_param_setup",
            root=self.root,
            log=self.log,
            citros=self,
            new=self.new,
            verbose=self.verbose,
            debug=self.debug,
            level=self.level + 1,
        )
        (self.root_citros / "parameter_setups").mkdir(parents=True, exist_ok=True)
        with importlib_resources.files(f"data.doc").joinpath(
            "parameter_setups/README.md"
        ) as md_file_path:
            shutil.copy2(md_file_path, self.root_citros / f"parameter_setups/README.md")

        # create simulation per launch file as default
        self.log.debug(f"creating .citros/simulations/*")
        (self.root_citros / "simulations").mkdir(parents=True, exist_ok=True)
        self._create_simulations()
        with importlib_resources.files(f"data.doc").joinpath(
            "simulations/README.md"
        ) as md_file_path:
            shutil.copy2(md_file_path, self.root_citros / f"simulations/README.md")

        self.log.debug(f"creating .citros/.gitignore")
        self._create_gitignore()

        self.log.debug(f"creating .citros/notebooks")
        (self.root_citros / "notebooks").mkdir(parents=True, exist_ok=True)
        with importlib_resources.files(f"data.doc").joinpath(
            "notebooks/README.md"
        ) as md_file_path:
            shutil.copy2(md_file_path, self.root_citros / f"notebooks/README.md")
        # TODO: copy some sample notebooks.

        self.log.debug(f"creating .citros/data")
        (self.root_citros / "data").mkdir(parents=True, exist_ok=True)
        with importlib_resources.files(f"data.doc").joinpath(
            "data/README.md"
        ) as md_file_path:
            shutil.copy2(md_file_path, self.root_citros / f"data/README.md")

        self.log.debug(f"creating .citros/reports")
        (self.root_citros / "reports").mkdir(parents=True, exist_ok=True)
        with importlib_resources.files(f"data.doc").joinpath(
            "reports/README.md"
        ) as md_file_path:
            shutil.copy2(md_file_path, self.root_citros / f"reports/README.md")

    #################
    ##### utils #####
    #################
    def _get_launches(self):
        """returns a list of launch objects

        Args:
            proj_json (Path): path to project.json file

        Returns:
            [{
                package: str,
                name: str
            }]: array of launch info
        """

        launch_info = []

        for package in self.citros.get("packages", []):
            for launch in package.get("launches", []):
                if "name" in launch:
                    launch_info.append(
                        {"package": package.get("name", ""), "name": launch["name"]}
                    )

        return launch_info

    def _create_simulations(self):
        launch_infos = self._get_launches()
        if not launch_infos:
            self.log.warning("No launch files found in user's project.")
            print(
                Panel.fit(
                    Padding(
                        "[yellow]No launch files found. [white]If you have launch files in your project, make sure they are of the form [green]*.launch.py ",
                        1,
                    ),
                    title="help",
                )
            )

            return

        # inspect(launch_infos)
        for launch in launch_infos:
            package_name = launch["package"]
            launch_file = launch["name"]

            # print("vova", self.root, f"simulation_{launch_file.split('.')[0]}.json")
            self.simulations.append(
                Simulation(
                    f"simulation_{launch_file.split('.')[0]}",
                    root=self.root,
                    new=self.new,
                    log=self.log,
                    citros=self,
                    package_name=package_name,
                    launch_file=launch_file,
                    verbose=self.verbose,
                    debug=self.debug,
                    level=self.level + 1,
                )
            )

    def _create_gitignore(self):
        if not Path(self.root, ".gitignore").exists():
            with open(Path(self.root, ".gitignore"), "w") as file:
                ignores = linesep.join(
                    [".citros/data/", ".citros/logs/"]
                )  # add more as needed.
                file.write(ignores)



import os
from typing import cast

from ..simulation import Simulation
from rich import print, inspect, print_json
from rich.rule import Rule
from rich.panel import Panel
from rich.padding import Padding
from rich.logging import RichHandler
from rich.console import Console
from rich.markdown import Markdown
from rich_argparse import RichHelpFormatter

from rich.traceback import install
from rich.logging import RichHandler
from rich import print, inspect, print_json
from rich.panel import Panel
from rich.padding import Padding

install()

from datetime import datetime


################################
# Entrypoint
################################
def generate_launch_description(simulation: Simulation, destination: str, events):
    """
    Generates a ROS2 LaunchDescription for a simulation run.

    Args:
        citros (Citros): An instance of Citros class which manages interaction with the CiTROS service.

    Returns:
        launch.LaunchDescription: A launch description that ROS2 can execute.
    """
    # running inside ROS workspace context.
    from ament_index_python.packages import get_package_share_directory
    from launch import LaunchDescription, Event
    from launch.actions import (
        EmitEvent,
        ExecuteProcess,
        IncludeLaunchDescription,
        DeclareLaunchArgument,
        OpaqueFunction,
        RegisterEventHandler,
        LogInfo,
        TimerAction,
    )
    from launch.launch_description_sources import PythonLaunchDescriptionSource
    from launch_ros.actions import Node
    from launch.substitutions import (
        LaunchConfiguration,
        LocalSubstitution,
        TextSubstitution,
    )
    from launch.event_handlers import (
        OnExecutionComplete,
        OnProcessExit,
        OnProcessIO,
        OnProcessStart,
        OnShutdown,
    )
    from launch.events import Shutdown, process
    from launch.actions import SetLaunchConfiguration

    events.init(
        tag="INIT",
        message="initializing launch",
        metadata=None,
    )

    ld = LaunchDescription([LogInfo(msg="CITROS launch file!")])

    # ld.add_action(SetLaunchConfiguration("simulation_name", simulation.name))
    # ld.add_action(SetLaunchConfiguration("simulation_run_dir", simulation_run_dir))
    # ld.add_action(SetLaunchConfiguration("batch_run_id", batch_run_id))
    # ld.add_action(SetLaunchConfiguration("simulation_run_id", simulation_run_id))

    # batch = simulation.batch.get_batch(batch_run_id, simulation_name)

    # override with simulation value
    timeout_from_client = simulation["timeout"]
    if float(timeout_from_client) > 1:
        timeout = str(timeout_from_client)
    ld.add_action(SetLaunchConfiguration("timeout", timeout))

    simulation.log.info(
        f"initializing simulation: {simulation.name} , timeout: {timeout}"
    )

    ################################
    # Arguments
    ################################
    # ld.add_action(
    #     DeclareLaunchArgument(
    #         "log_level",
    #         default_value=["info"],
    #         description="Logging level",
    #     )
    # )

    # ld.add_action(
    #     DeclareLaunchArgument(
    #         "simulation_name",
    #         description="simulation name",
    #     )
    # )

    # ld.add_action(
    #     DeclareLaunchArgument(
    #         "simulation_run_dir",
    #         description="simulation metadata directory",
    #     )
    # )

    # ld.add_action(
    #     DeclareLaunchArgument(
    #         "batch_run_id",
    #         description=("Batch Run id"),
    #     )
    # )

    # ld.add_action(
    #     DeclareLaunchArgument(
    #         "simulation_run_id",
    #         description=(
    #             "Simulation run id, as part of [sequence]/[simulation.repeats]"
    #         ),
    #     )
    # )

    ld.add_action(
        DeclareLaunchArgument(
            "timeout",
            description=("The timeout for the simulation [sec]"),
            default_value=str(60 * 60 * 24 * 7),
        )
    )

    ################################
    # RECORDING BAG Proccess
    ################################

    # the simulation run directory is created ad-hoc before the call to generate_launch_description.
    bag_folder = destination
    bag_cmd = ["ros2", "bag", "record", "-a", "-o", f"{bag_folder}/bags"]

    mcap = simulation["storage_type"] == "MCAP"
    if mcap:
        bag_cmd.append("-s")
        bag_cmd.append("mcap")

    record_proccess = ExecuteProcess(
        cmd=bag_cmd, 
        name="citros_bag_recorder", 
        output="screen",
        log_cmd=True,
    )
    ld.add_action(record_proccess)

    # ################################
    # # STD LOG on ros events.
    # ################################
    # # Setup a custom event handler for all stdout/stderr from processes.
    # # Later, this will be a configurable, but always present, extension to the LaunchService.
    # def on_output(event: Event) -> None:
    #     for line in event.text.decode().splitlines():
    #         # log files will be written to ROS_LOGS_DIR rather than CLI_LOGS _DIR
    #         simulation.log.info(f"[ROS][{cast(process.ProcessIO, event).process_name}]{line}")

    # ld.add_action(RegisterEventHandler(
    #     OnProcessIO(
    #             on_stdout=on_output,
    #             on_stderr=on_output,
    #         )
    #     )
    # )

    ################################
    # User launch file
    ################################

    def launch_setup(context, *args, **kwargs):
        simulation.log.debug("launch_setup")

        # simulation_name = LaunchConfiguration("simulation_name").perform(context)
        # simulation_run_dir = LaunchConfiguration("simulation_run_dir").perform(context)

        # batch_run_id = LaunchConfiguration("batch_run_id").perform(context)
        # simulation_run_id = LaunchConfiguration("simulation_run_id").perform(context)

        # config
        # config = simulation.params.init_params(
        #     # simulation_name, 
        #     simulation_run_dir, 
        #     # int(simulation_run_id)
        # )
        # send event with the config to CiTROS
        # events.starting(
        #     batch_run_id=batch_run_id,
        #     sid=simulation_run_id,
        #     tag="CONFIG",
        #     message="updated config",
        #     metadata=config,
        # )

        # launch
        launch_name = simulation["launch"]["file"]
        launch_package = simulation["launch"]["package"]

        client_launch = IncludeLaunchDescription(
            PythonLaunchDescriptionSource(
                [
                    os.path.join(get_package_share_directory(launch_package), "launch"),
                    f"/{launch_name}",
                ]
            ),
            launch_arguments={}.items(),            
        )

        simulation.log.info(
            f"Starting clients launch package:{launch_package} launch:{launch_name}"
        )
        events.running(
            # batch_run_id=batch_run_id,
            # sid=simulation_run_id,
            tag="LAUNCH",
            message="Starting clients launch",
            metadata=None,
        )

        return [client_launch]

    ld.add_action(OpaqueFunction(function=launch_setup))

    ################################
    # Timeout Events
    ################################
    def handle_timeout(context, *args, **kwargs):
        simulation.log.debug("handle_timeout")
        # batch_run_id = LaunchConfiguration("batch_run_id").perform(context)
        # simulation_run_id = LaunchConfiguration("simulation_run_id").perform(context)
        timeout = LaunchConfiguration("timeout").perform(context)
        events.terminating(
            # batch_run_id,
            # simulation_run_id,
            tag="TIMEOUT",
            message=f"Reached timeout of: { timeout } sec",
            metadata=None,
        )

    ld.add_action(
        TimerAction(
            period=LaunchConfiguration("timeout"),
            actions=[
                LogInfo(msg="---------TIMEOUT---------"),
                OpaqueFunction(function=handle_timeout),
                EmitEvent(event=Shutdown(reason="TIMEOUT")),
            ],
        )
    )

    ################################
    # Exit
    ################################
    def handle_shutdown(context, *args, **kwargs):
        simulation.log.debug("handle_shutdown")
        # batch_run_id = LaunchConfiguration("batch_run_id").perform(context)
        # simulation_run_id = LaunchConfiguration("simulation_run_id").perform(context)
        reason = LocalSubstitution("event.reason").perform(context)
        events.terminating(
            # batch_run_id,
            # simulation_run_id,
            tag="SHUTDOWN",
            message=reason,
            metadata=None,
        )

    ld.add_action(
        RegisterEventHandler(
            OnShutdown(
                on_shutdown=[
                    OpaqueFunction(function=handle_shutdown),
                    LogInfo(
                        msg=[
                            "Launch was asked to shutdown: ",
                            LocalSubstitution("event.reason"),
                        ]
                    ),
                ]
            )
        )
    )

    ld.add_action(
        RegisterEventHandler(
            OnProcessExit(
                target_action=record_proccess,
                on_exit=[
                    LogInfo(msg=(f"The simulation has finished.")),
                    EmitEvent(event=Shutdown(reason="Finished")),
                ],
            )
        )
    )

    return ld



#!/usr/bin/env python3
# ==============================================
#  ██████╗██╗████████╗██████╗  ██████╗ ███████╗
# ██╔════╝██║╚══██╔══╝██╔══██╗██╔═══██╗██╔════╝
# ██║     ██║   ██║   ██████╔╝██║   ██║███████╗
# ██║     ██║   ██║   ██╔══██╗██║   ██║╚════██║
# ╚██████╗██║   ██║   ██║  ██║╚██████╔╝███████║
#  ╚═════╝╚═╝   ╚═╝   ╚═╝  ╚═╝ ╚═════╝ ╚══════╝
# ==============================================
from setuptools import setup, find_packages
import citros_meta

try:
    import sys
    from semantic_release import setup_hook

    setup_hook(sys.argv)
except ImportError:
    pass


setup(
    name=citros_meta.__title__,
    version=citros_meta.__version__,
    author=citros_meta.__author__,
    author_email=citros_meta.__author_email__,
    packages=find_packages(),
    package_data={"": ["*.json", "*.sh", "*.py", "*.md", ".citrosignore"]},
    # scripts=[
    #     'bin/citros',
    # ],
    entry_points={
        "console_scripts": [
            "citros = bin.cli:main",
        ],
    },
    url=citros_meta.__url__,
    license=citros_meta.__license__,
    description="A cli entrypoint for the citros system.",
    long_description_content_type="text/markdown",
    long_description=open("pypi.md").read(),
    # TODO[critical] - import from requierments.txt
    install_requires=[
        "ansicolors",
        "gql",
        "graphql-core",
        "requests",
        "rosdep",
        "python-decouple",
        "requests_toolbelt",
        "soupsieve",
        "bs4",
        "zipp",
        "pyjwt",
        "psycopg2-binary",
        "urllib3",
        "InquirerPy",
        "GitPython",
        "jsonschema",
        "cryptography",
        "numpy",
        "psutil",
        "cmakeast",
        "opentelemetry-api",
        "opentelemetry-sdk",
        "opentelemetry-exporter-otlp-proto-grpc",
        "pydantic==1.10.12",
        "python-on-whales",
        "importlib-resources",
        "citros-data-analysis",
    ],
    py_modules=["citros", "citros_meta", "data"],
    project_urls={
        "Documentation": "http://citros.io/doc/docs_cli",
        "Source Code": "https://github.com/lulav/citros_cli",
        "Bug Tracker": "https://github.com/lulav/citros_cli/issues",
    },
)
